;=========================================================================
;
;            ATCLKFIX
;
;            CLOCK$ Device driver.  For use on AT-type machines only.
;            Like the driver in PC DOS v2.00 and v2.10, except the `midnight'
;            bug is fixed and the real-time clock date is read.
;
;=========================================================================

;******** Prototype for Request Header
REQ_HDR      STRUC
RH_LEN       DB      ?                   ; Offset 0
RH_UNIT_CODE DB      ?                   ; Offset 1
RH_CMD_CODE  DB      ?                   ; Offset 2
RH_STATUS    DW      ?                   ; Offset 3
RH_RESERVED  DB      8 DUP(?)            ; Offset 5
RH_MEDIA     DB      ?                   ; Offset 13
RH_ADDR_OFF  DW      ?                   ; Offset 14
RH_ADDR_SEG  DW      ?                   ; Offset 16
RH_BYTPSEC   DW      ?                   ; Offset 18
REQ_HDR      ENDS

;******** Prototype for Clock Data
CLOCK_DATA   STRUC
CD_DAYS      DW      ?
CD_HR_MIN    DW      ?
CD_SECONDS   DW      ?
CLOCK_DATA   ENDS

;******** Equates
ZERO         EQU     0
SIXTY        EQU     60
HUNDRED      EQU     100
CON59        EQU     59659               ; 59659 / (5*64K) ticks/100th-sec
CR           EQU     0DH
LF           EQU     0AH

;==========================================================================
;
;            Executable Code
;
;==========================================================================

DEVSEG       SEGMENT PARA PUBLIC 'CODE'

             ASSUME  CS:DEVSEG, DS:DEVSEG
             ORG     ZERO

DEVPROC      PROC    FAR

;==========================================================================
;            Data Storage Area
;==========================================================================

;******** Header for DOS Device Drivers
NEXT_DEVIP   DW      -1                  ; fake pointer to ..
NEXT_DEVCS   DW      -1                  ; .. next device driver
ATTRIBUTE    DW      8008H               ; Char device and current clock
STRAT_PTR    DW      OFFSET STRATEGY
INTRP_PTR    DW      OFFSET COMMANDS
DEV_NAME     DB      'CLOCK$  '
;
;******** Pointer to DOS Request Header
REQ_HDR_PTR  LABEL   DWORD
 RHP_OFF     DW      ZERO
 RHP_SEG     DW      ZERO
;
;******** Numerical data
DAYCOUNT     DW      ZERO                ; 0 corresponds to 01 Jan 80
NUMTICKSHI   DW      ZERO
;
;******** Jump Table for Interrupt Routine Functions
FN_TABLE     DW      INITIALIZE          ;  0: Initialization Procedure
             DW      ERROR_EXIT          ;  1: Media Check Procedure
             DW      ERROR_EXIT          ;  2: Build BIOS Parameter Block
             DW      ERROR_EXIT          ;  3: IOCTL Input
             DW      READ_DEV            ;  4: Read from device
             DW      READ_EXIT           ;  5: Non-destructive no-wait read
             DW      GOOD_EXIT           ;  6: Return status of input device
             DW      GOOD_EXIT           ;  7: Flush input buffer
             DW      WRITE_DEV           ;  8: Write to device
             DW      WRITE_DEV           ;  9: Write and verify
             DW      GOOD_EXIT           ; 10: Return status of output device
             DW      GOOD_EXIT           ; 11: Flush output buffer

;==========================================================================
;            Device Strategy Routine
;            External Entry Point
;==========================================================================

             ASSUME  CS:DEVSEG, DS:NOTHING, ES:NOTHING

STRATEGY     PROC    FAR
             MOV     CS:RHP_OFF,BX
             MOV     CS:RHP_SEG,ES
             RETF
STRATEGY     ENDP


;==========================================================================
;            Interrupt Handler Routine
;            External Entry Point
;==========================================================================

             ASSUME  CS:DEVSEG, DS:NOTHING, ES:NOTHING

COMMANDS     PROC    FAR

;=======================================
;======== Main Routine
;=======================================
;
;******** Common setup for all functions
;-------- Clear direction
             CLD
;-------- Push registers
             PUSH    AX
             PUSH    BX
             PUSH    CX
             PUSH    DX
             PUSH    SI
             PUSH    DI
             PUSH    BP
             PUSH    DS
             PUSH    ES
;-------- Make DS:BX point at Request Header
             LDS     BX,REQ_HDR_PTR
;
;******** Processing based on Request Header command code
;-------- Place the command code in AL
             MOV     AL,[BX].RH_CMD_CODE
;-------- Branch and exit if command code is greater than 11
             CMP     AL,0BH
             JA      ERROR_EXIT
;-------- Load CX from Request Header `bytes per sector' field
             MOV     CX,[BX].RH_BYTPSEC
;-------- Load ES:DI from Request Header `address' field
             LES     DI,DWORD PTR [BX].RH_ADDR_OFF
;-------- Point DS at this segment
             PUSH    CS
             POP     DS
             ASSUME  DS:DEVSEG
;-------- Jump to routine called for in command code
             MOV     SI,OFFSET FN_TABLE
             MOV     AH,ZERO
             SHL     AX,1
             ADD     SI,AX
             JMP     WORD PTR [SI]
;
;******** Common exit from Interrupt Handler routine
;-------- Assign result status and branch to STOR_STATUS
READ_EXIT:   MOV     AX,0300H            ; Code for `done and busy'
             JMP     SHORT STOR_STATUS
;-------- Assign bad result status and branch to STOR_STATUS
ERROR_EXIT:  MOV     AX,8103H            ; Code for `unknown command'
             JMP     SHORT STOR_STATUS
;-------- Assign result status
GOOD_EXIT:   MOV     AX,100H             ; Code for `done'
;-------- Store result status in Request Header
STOR_STATUS: LDS     BX,REQ_HDR_PTR
             MOV     [BX].RH_STATUS,AX
;-------- Restore registers
             POP     ES
             POP     DS
             POP     BP
             POP     DI
             POP     SI
             POP     DX
             POP     CX
             POP     BX
             POP     AX
;-------- Return
             RETF


;=======================================
;======== `Write to Device' Processing routine
;=======================================
;
;         Input: 6 byte block provided by DOS and pointed to by
;                RH_ADDR_SEG:RH_ADDR_OFF (which has already been
;                placed in ES:DI).  The 6 byte block has the format
;                of CLOCK_DATA.
;         Processing: Saves input days info in location DAYCOUNT.
;                     Converts input hours/minutes/seconds info to the
;                     number of 55 msec ticks since midnight, as follows:
;                        a) Convert hr/min/sec to hundredths-of-seconds
;                        b) Multiply by 59659 / (5*64K) = 0.182
;         Output: Writes the number of 55 msec ticks to BIOS Data Area
;                 (at 0040:006C) using BIOS Interrupt 1AH.
;
WRITE_DEV    LABEL   NEAR
;
;******** Move data from DOS to this program
;-------- Place days since 1/1/80 in DAYCOUNT
             MOV     AX,ES:[DI].CD_DAYS
             MOV     DAYCOUNT,AX
;-------- Place hours in CH, minutes in CL
             MOV     CX,ES:[DI].CD_HR_MIN
;-------- Place hundredth of seconds in DL, seconds in DH
             MOV     DX,ES:[DI].CD_SECONDS
;
;******** Convert hr/min/sec/100ths to hundredths-of-seconds, and put in DX,CX
;-------- Convert from hr:min to total minutes, and place in AX
             MOV     AL,SIXTY
             MUL     CH
             MOV     CH,ZERO
             ADD     AX,CX
;-------- Convert total minutes to hundredth of seconds, and place in DX,CX
             MOV     CX,SIXTY*HUNDRED    ; Hundredths-of-seconds per minute
             MOV     BX,DX               ; Stores CD_SECONDS in BX
             MUL     CX
             MOV     CX,AX
;-------- Convert seconds to hundredths-of-seconds and add to CX
             MOV     AL,HUNDRED          ; Hundredths-of-seconds per second
             MUL     BH
             ADD     CX,AX
             ADC     DX,ZERO             ; Take care of the carry
;-------- Add in the hundredths-of-seconds
             MOV     BH,ZERO
             ADD     CX,BX
             ADC     DX,ZERO
;
;******** Multiply DX,CX by 59,659 and discard lowest word of three-word result
;-------- Place hundredths-of-seconds in CX,AX and free DX
             XCHG    DX,AX
             XCHG    CX,AX
;-------- Multiply hundredths-of-seconds low-word in AX by 59,659
             MOV     BX,CON59
             MUL     BX
;-------- Place results in CX,DX, and hundredths-of-second high-word in AX
             XCHG    DX,CX
             XCHG    DX,AX
;-------- Multiply hundredths-of-seconds high-word by 59,659, over-writing DX
             MUL     BX
;-------- Add in most significant half of low-word results
             ADD     AX,CX
             ADC     DX,ZERO
;-------- Place result in AX,DX -- new high- and low-words
             XCHG    DX,AX
;
;******** Divide AX,DX by 5, and place result in CX,DX
;-------- Divide AX by 5 (byte operation)
             MOV     BX,5
             DIV     BL
;-------- Save quotient in CX (high-word of tick count)
             MOV     CL,AL
             MOV     CH,ZERO
;-------- Place remainder in DX, low-word in AX
             MOV     AL,AH
             CBW
             XCHG    DX,AX
;-------- Divide DX,AX by 5 (word operation)
             DIV     BX
;-------- Place quotient in DX (low-word of tick count), over-writing remainder
             MOV     DX,AX
;
;******** Save result and exit
;-------- Save the high-word part of tick count
             MOV     NUMTICKSHI,CX
;-------- Invoke BIOS routine to set the tick count (now in CX,DX)
             MOV     AH,01H
             INT     1AH
;-------- Jump to common exit
             JMP     GOOD_EXIT


;=======================================
;======== Read Processing routine
;=======================================
;
;         Input: Two-word block (provided by BIOS Interrupt 1AH) containing
;                the number of 55 msec ticks currently in the BIOS Data
;                Area at 0040:006C
;         Processing: Updates internal variable DAYCOUNT if midnight passed.
;                     Converts number of 55 msec clock ticks to hours/minutes/
;                     seconds/100ths info as follows:
;                        a) Multiply by (5*64K) / 59659 = 1 / 0.182
;                        b) Convert hundredths-of-seconds to hr/min/sec/100ths
;         Output: Information is placed in 6 byte block pointed to by
;                 RH_ADDR_SEG:RH_ADDR_OFF -- block has the format of
;                 CLOCK_DATA.
;
READ_DEV     LABEL   NEAR
;
;******** Get tick count into CX,DX, and increment DAYCOUNT if midnight passed
;-------- Invoke BIOS routine to read the tick count into CX,DX
             XOR     AX,AX
             INT     1AH
;-------- If NUMTICKSHI has decreased since last saved, then increment DAYCOUNT
             CMP     CX,NUMTICKSHI
             JNB     READ_1
             INC     DAYCOUNT
;-------- Save the high-word part of tick count
READ_1:      MOV     NUMTICKSHI,CX
;
;******** Multiply CX,DX by 5 and put result in DX,AX
;-------- Save tick count in AX,BX
             MOV     AX,CX
             MOV     BX,DX
;-------- Multiply CX,DX by 2 using shifts
             SHL     DX,1
             RCL     CX,1
;-------- Multiply CX,DX by 2 using shifts
             SHL     DX,1
             RCL     CX,1
;-------- Add AX,BX to CX,DX and put result in AX,DX
             ADD     DX,BX
             ADC     AX,CX
;-------- Exchange AX and DX -- result in DX,AX
             XCHG    DX,AX
;
;******** Convert to hundredths-of-seconds by multiplying by 64K / 59659
;-------- Divide DX,AX by 59,659 (word operation)
             MOV     CX,CON59
             DIV     CX
;-------- Place quotient in BX
             MOV     BX,AX
;-------- Append word to remainder and divide by 59,659 again
             XOR     AX,AX
             DIV     CX
;
;******** Find number of 100th-sec and total seconds since midnight
;-------- Place hundredths-of-seconds in DX,AX
             MOV     DX,BX
;-------- Divide DX,AX by 200 so quotient (total-seconds/2) fits in one word
             MOV     CX,200
             DIV     CX
;-------- Branch around if remainder is less than 100
             CMP     DL,HUNDRED
             JB      READ_2              ; Carry Flag now 1
;-------- Subtract 100 from remainder (sets CF to 0)
             SUB     DL,HUNDRED
;-------- Complement carry flag (set it to 1 if remainder was 100 or greater)
READ_2:      CMC
;-------- Place 100th-sec byte into BL
             MOV     BL,DL
;
;******** Convert total seconds since midnight to hours/minutes/seconds
;-------- Place total seconds since midnight in DX,AX
             RCL     AX,1                ; Rotate thru carry
             MOV     DL,ZERO
             RCL     DX,1                ; Rotate thru carry
;-------- Divide total seconds by 60
             MOV     CX,SIXTY
             DIV     CX
;-------- Place remainder (seconds byte) in BH
             MOV     BH,DL
;-------- Divide quotient (total minutes since midnight) by 60
             DIV     CL
;-------- Put hours byte in AH, minutes byte in AL
             XCHG    AH,AL
;
;******** Put result in 6-byte block pointed to by ES:DI, and exit
;-------- Push hour/minute info on stack
             PUSH    AX
;-------- Get day info from internal variable and put in block
             MOV     AX,DAYCOUNT
             STOSW                       ; Store ax to es:[di]
;-------- Pop hour/minute info and put in block
             POP     AX
             STOSW                       ; Store ax to es:[di]
;-------- Get seconds info and put in block
             MOV     AX,BX
             STOSW                       ; Store ax to es:[di]
;-------- Jump to common exit
             JMP     GOOD_EXIT


;=======================================
;======== Initialization Routine
;=======================================
;
DAYS_MONTH   DW      0,31,59,90,120,151,181,212,243,273,304,334
MSG$         DB      CR,LF,'ATCLKFIX Installed',CR,LF,CR,LF,0
;
INITIALIZE   LABEL   NEAR
;
;******** Point DOS at end of resident code
;-------- Make ES:BX point at Request Header
             LES     BX,REQ_HDR_PTR
;-------- Place end of device address in Request Header `address' field
             MOV     ES:[BX].RH_ADDR_OFF,OFFSET DAYS_MONTH
             MOV     ES:[BX].RH_ADDR_SEG,CS
;
;******** Read real-time clock date info and convert from BCD to binary
;-------- Do read (CH=year high digits, CL=year low digits, DH=month, DL=day)
             MOV     AH,04H
             INT     1AH
;-------- Convert month
             MOV     AL,DH
             CALL    BCD_TO_BIN
             MOV     DH,AL
             DEC     DH
;-------- Convert day
             MOV     AL,DL
             CALL    BCD_TO_BIN
             MOV     DL,AL
             DEC     DL
;-------- Convert last two digits of year
             MOV     AL,CL
             CALL    BCD_TO_BIN
             MOV     CL,AL
;-------- Convert first two digits of year and place in AX
             MOV     AL,CH
             CALL    BCD_TO_BIN
             MOV     AH,HUNDRED
             MUL     AH
;-------- Add two halves of year together and place in CX
             XOR     CH,CH
             ADD     CX,AX
;
;******** Convert from year/month/day form to number-of-days since 1/1/80
;-------- Put number of days in this (assumed non-leap) year in BX
             XOR     BX,BX
             MOV     BL,DL
             XOR     AX,AX
             MOV     AL,DH
             SHL     AX,1
             MOV     SI,OFFSET DAYS_MONTH
             ADD     SI,AX
             ADD     BX,WORD PTR [SI]
;-------- Account for current leap year
             TEST    CX,0003H            ; Test for leap year
             JNZ     INIT_1              ; Loop around if not
             CMP     DX,0201H            ; Test against March 1
             JB      INIT_1              ; Loop around if less
             INC     BX                  ; Increment otherwise
;-------- Account for previous leap years
INIT_1:      MOV     AX,CX
             SUB     AX,1977
             SHR     AX,1                ; Divide ..
             SHR     AX,1                ; .. by 4
             ADD     BX,AX
;-------- Account for 365 days per previous year
             MOV     AX,365
             SUB     CX,1980
             MUL     CX
             ADD     AX,BX
;
;******** Save number-of-days and branch to exit
;-------- Save number-of-days in internal storage
             MOV     DAYCOUNT,AX
;-------- Display message
             MOV     SI,OFFSET MSG$
             CALL    DISP_MSG
;-------- Jump to common exit
             JMP     GOOD_EXIT


;======== Subroutine BCD_TO_BIN
;
;         Input: Integer between 0 and 99, stored in AL, coded in BCD
;                format (viz, a decimal integer in each nibble)
;         Output: Same integer, stored in AL, coded in binary format
;
BCD_TO_BIN   PROC    NEAR
;-------- BL holds low nibble, AL holds high nibble
             MOV     BL,AL
             AND     BL,0FH
             AND     AL,0F0H
;-------- Multiply high nibble by 10
             ROR     AL,1                ; Rotate
             MOV     AH,AL
             ROR     AL,1                ; Rotate
             ROR     AL,1                ; Rotate
             ADD     AL,AH               ; Low nibble*8 + low nibble*2
;-------- Add low nibble to result
             ADD     AL,BL
;-------- Return to caller
             RETN
BCD_TO_BIN   ENDP


;======== Subroutine DISP_MSG
;
;         Display message on screen.
;         Input:  SI = offset (within this segment) of message start
;                 Assumes DS points at this segment
;
DISP_MSG     PROC    NEAR
;-------- Fetch and display one character
NXTLTR:      LODSB                        ; Load char into AL
             AND     AL,7FH               ; Test for zero
             JZ      DMRET                ; Jump and return if zero
             MOV     AH,0EH               ; WRITE_TTY service
             MOV     BX,0007H             ; BH=page, BL=foreground
             INT     10H                  ; Invoke BIOS routine
             JMP     SHORT   NXTLTR       ; Repeat for next char
;-------- Return to caller
DMRET:       RETN
DISP_MSG     ENDP


COMMANDS     ENDP

DEVPROC      ENDP

DEVSEG       ENDS

             END
